"""
This module contains hashing functions for Python objects. It uses
functools.singledispatch to allow specialized implementations based on type.
Singledispatch automatically applies the most specific implementation

This module houses implementations for the Python standard library. Supporting
all types is considerable endeavor, so we'll add support as types are requested
by users.

Otherwise, 3rd party types can be supported via the `h_databackends` module.
This registers abstract types that can be checked without having to import the
3rd party library. For instance, there are implementations for pandas.DataFrame
and polars.DataFrame despite these libraries not being imported here.

IMPORTANT all container types that make a recursive call to `hash_value` or a specific
implementation should pass the `depth` parameter to prevent `RecursionError`.
"""

import base64
import datetime
import functools
import hashlib
import logging
import sys
from collections.abc import Mapping, Sequence, Set
from typing import Dict

from hamilton.experimental import h_databackends

# NoneType is introduced in Python 3.10
try:
    from types import NoneType
except ImportError:
    NoneType = type(None)


logger = logging.getLogger("hamilton.caching")


MAX_DEPTH = 6
UNHASHABLE = "<unhashable>"
NONE_HASH = "<none>"


def set_max_depth(depth: int) -> None:
    """Set the maximum recursion depth for fingerprinting non-supported types.

    :param depth: The maximum depth for fingerprinting.
    """
    global MAX_DEPTH
    MAX_DEPTH = depth


def _compact_hash(digest: bytes) -> str:
    """Compact the hash to a string that's safe to pass around.

    NOTE this is particularly relevant for the Hamilton UI and
    passing hashes/fingerprints through web services.
    """
    return base64.urlsafe_b64encode(digest).decode()


@functools.singledispatch
def hash_value(obj, *args, depth=0, **kwargs) -> str:
    """Fingerprinting strategy that computes a hash of the
    full Python object.

    The default case hashes the `__dict__` attribute of the
    object (recursive).
    """
    if depth > MAX_DEPTH:
        return UNHASHABLE

    if hasattr(obj, "__dict__"):
        return hash_value(obj.__dict__, depth=depth + 1)

    # check if the object comes from a module part of the standard library
    # if it's the case, hash it's __repr__(), which is a string representation of the object
    # __repr__() from the standard library should be well-formed and offer a reliable basis
    # for fingerprinting.
    # for example, this will catch: pathlib.Path, enum.Enum, argparse.Namespace
    elif getattr(obj, "__module__", False):
        if obj.__module__.partition(".")[0] in sys.builtin_module_names:
            return hash_repr(obj, depth=depth)

    # cover the datetime module, which doesn't have a __module__ attribute
    elif type(obj) in vars(datetime).values():
        return hash_repr(obj, depth=depth)

    return UNHASHABLE


@hash_value.register(NoneType)
def hash_none(obj, *args, **kwargs) -> str:
    """Hash for None is <none>

    Primitive type returns a hash and doesn't have to handle depth.
    """
    return NONE_HASH


def hash_repr(obj, *args, **kwargs) -> str:
    """Use the built-in repr() to get a string representation of the object
    and hash it.

    While `.__repr__()` might not be implemented for all classes, the function
    `repr()` will handle it, along with exceptions, to always return a value.

    Primitive type returns a hash and doesn't have to handle depth.
    """
    return hash_primitive(repr(obj))


# we need to use explicit multiple registration because older Python
# versions don't support type annotations with Union types
@hash_value.register(str)
@hash_value.register(int)
@hash_value.register(float)
@hash_value.register(bool)
def hash_primitive(obj, *args, **kwargs) -> str:
    """Convert the primitive to a string and hash it

    Primitive type returns a hash and doesn't have to handle depth.
    """
    hash_object = hashlib.md5(str(obj).encode())
    return _compact_hash(hash_object.digest())


@hash_value.register(bytes)
def hash_bytes(obj, *args, **kwargs) -> str:
    """Convert the primitive to a string and hash it

    Primitive type returns a hash and doesn't have to handle depth.
    """
    hash_object = hashlib.md5(obj)
    return _compact_hash(hash_object.digest())


@hash_value.register(Sequence)
def hash_sequence(obj, *args, depth: int = 0, **kwargs) -> str:
    """Hash each object of the sequence.

    Orders matters for the hash since orders matters in a sequence.
    """
    hash_object = hashlib.sha224()
    for elem in obj:
        hash_object.update(hash_value(elem, depth=depth + 1).encode())

    return _compact_hash(hash_object.digest())


def hash_unordered_mapping(obj, *args, depth: int = 0, **kwargs) -> str:
    """

    When hashing an unordered mapping, the two following dict have the same hash.

    .. code-block:: python

        foo = {"key": 3, "key2": 13}
        bar = {"key2": 13, "key": 3}

        hash_mapping(foo) == hash_mapping(bar)
    """

    hashed_mapping: Dict[str, str] = {}
    for key, value in obj.items():
        hashed_mapping[hash_value(key, depth=depth + 1)] = hash_value(value, depth=depth + 1)

    hash_object = hashlib.sha224()
    for key, value in sorted(hashed_mapping.items()):
        hash_object.update(key.encode())
        hash_object.update(value.encode())

    return _compact_hash(hash_object.digest())


@hash_value.register(Mapping)
def hash_mapping(obj, *, ignore_order: bool = True, depth: int = 0, **kwargs) -> str:
    """Hash each key then its value.

    The mapping is always sorted first because order shouldn't matter
    in a mapping.

    NOTE Since Python 3.7, dictionary store insertion order. However, this
    function assumes that they key order doesn't matter to uniquely identify
    the dictionary.

    .. code-block:: python

        foo = {"key": 3, "key2": 13}
        bar = {"key2": 13, "key": 3}

        hash_mapping(foo) == hash_mapping(bar)

    """
    if ignore_order:
        # use the same depth because we're simply dispatching to another implementation
        return hash_unordered_mapping(obj, depth=depth)

    hash_object = hashlib.sha224()
    for key, value in obj.items():
        hash_object.update(hash_value(key, depth=depth + 1).encode())
        hash_object.update(hash_value(value, depth=depth + 1).encode())

    return _compact_hash(hash_object.digest())


@hash_value.register(Set)
def hash_set(obj, *args, depth: int = 0, **kwargs) -> str:
    """Hash each element of the set, then sort hashes, and
    create a hash of hashes.

    For the same objects in the set, the hashes will be the
    same.
    """
    hashes = [hash_value(elem, depth=depth + 1) for elem in obj]
    sorted_hashes = sorted(hashes)

    hash_object = hashlib.sha224()
    for hash in sorted_hashes:
        hash_object.update(hash.encode())

    return _compact_hash(hash_object.digest())


@hash_value.register(h_databackends.AbstractPandasDataFrame)
@hash_value.register(h_databackends.AbstractPandasColumn)
def hash_pandas_obj(obj, *args, depth: int = 0, **kwargs) -> str:
    """Convert a pandas dataframe, series, or index to
    a dictionary of {index: row_hash} then hash it.

    Given the hashing for mappings, the physical ordering or rows doesn't matter.
    For example, if the index is a date, the hash will represent the {date: row_hash},
    and won't preserve how dates were ordered in the DataFrame.
    """
    from pandas.util import hash_pandas_object

    hash_per_row = hash_pandas_object(obj)
    return hash_mapping(hash_per_row.to_dict(), ignore_order=False, depth=depth + 1)


@hash_value.register(h_databackends.AbstractPolarsDataFrame)
def hash_polars_dataframe(obj, *args, depth: int = 0, **kwargs) -> str:
    """Convert a polars dataframe, series, or index to
    a list of hashes then hash it.
    """
    hash_per_row = obj.hash_rows()
    return hash_sequence(hash_per_row.to_list(), depth=depth + 1)


@hash_value.register(h_databackends.AbstractPolarsColumn)
def hash_polars_column(obj, *args, depth: int = 0, **kwargs) -> str:
    """Promote the single Series to a dataframe and hash it"""
    # use the same depth because we're simply dispatching to another implementation
    return hash_polars_dataframe(obj.to_frame(), depth=depth)


@hash_value.register(h_databackends.AbstractNumpyArray)
def hash_numpy_array(obj, *args, depth: int = 0, **kwargs) -> str:
    """Get the bytes representation of the array raw data and hash it.

    Might not be ideal because different higher-level numpy objects could have
    the same underlying array representation (e.g., masked arrays).
    Unsure, but it's an area to investigate.
    """
    # use the same depth because we're simply dispatching to another implementation
    return hash_bytes(obj.tobytes(), depth=depth)



from typing import Any

try:
    import dask.dataframe as dd
except ImportError as e:
    raise NotImplementedError("Dask is not installed.") from e

from hamilton import registry

DATAFRAME_TYPE = dd.DataFrame
COLUMN_TYPE = dd.Series


@registry.get_column.register(dd.DataFrame)
def get_column_dask(df: dd.DataFrame, column_name: str) -> dd.Series:
    return df[column_name]


@registry.fill_with_scalar.register(dd.DataFrame)
def fill_with_scalar_dask(df: dd.DataFrame, column_name: str, value: Any) -> dd.DataFrame:
    df[column_name] = value
    return df


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("dask", DATAFRAME_TYPE, COLUMN_TYPE)


register_types()



import pathlib


def get_directory_size(directory: str) -> float:
    """Get the size of the content of a directory in bytes."""
    total_size = 0
    for p in pathlib.Path(directory).rglob("*"):
        if p.is_file():
            total_size += p.stat().st_size

    return total_size


def readable_bytes_size(n_bytes: float) -> str:
    """Convert a number of bytes to a human-readable unit."""
    labels = ["B", "KB", "MB", "GB", "TB"]
    exponent = 0

    while n_bytes > 1024.0:
        n_bytes /= 1024.0
        exponent += 1

    return f"{n_bytes:.2f} {labels[exponent]}"



import dataclasses
import functools
import logging
import types
from typing import Callable, Optional, Tuple, Union

from hamilton import version

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Version:
    major: int
    minor: int
    patch: int

    def __gt__(self, other: "Version"):
        return (self.major, self.minor, self.patch) > (other.major, other.minor, other.patch)

    @staticmethod
    def from_version_tuple(version_tuple: Tuple[Union[int, str], ...]) -> "Version":
        version_ = version_tuple
        if len(version_) > 3:  # This means we have an RC
            version_ = version_tuple[0:3]  # Then let's ignore it
        return Version(*version_)  # TODO, add some validation

    @staticmethod
    def current() -> "Version":
        current_version = version.VERSION
        return Version.from_version_tuple(current_version)

    def __repr__(self):
        return ".".join(map(str, [self.major, self.minor, self.patch]))


CURRENT_VERSION = Version.current()


class DeprecationError(Exception):
    def raise_(self):
        raise self


@dataclasses.dataclass
class deprecated:
    """Deprecation decorator -- use judiciously! For example:
    @deprecate(
        warn_starting=(1,10,0)
        fail_starting=(2,0,0),
        use_this=parameterize_values,
        explanation='We have redefined the parameterization decorators to consist of `parametrize`, `parametrize_inputs`, and `parametrize_values`
        migration_guide="https://github.com/dagworks-inc/hamilton/..."
    )
    class parameterized(...):
       ...
    Note this locks into a future contract (although it *can* be changed), so if you promise to deprecate something by X.0, then do it!

    """

    warn_starting: Union[Tuple[int, int, int], Version]
    fail_starting: Union[Tuple[int, int, int], Version]
    use_this: Optional[
        Callable
    ]  # If this is None, it means this functionality is no longer supported.
    explanation: str
    migration_guide: Optional[
        str
    ]  # If this is None, this means that the use_this is a drop in replacement
    current_version: Union[Tuple[int, int, int], Version] = dataclasses.field(
        default_factory=lambda: CURRENT_VERSION
    )
    warn_action: Callable[[str], None] = dataclasses.field(default=logger.warning)
    fail_action: Callable[[str], None] = dataclasses.field(
        default=lambda message: DeprecationError(message).raise_()
    )

    @staticmethod
    def _raise_failure(message: str):
        raise DeprecationError(message)

    @staticmethod
    def _ensure_version_type(version_spec: Union[Tuple[int, int, int], Version]) -> Version:
        if isinstance(version_spec, tuple):
            return Version(*version_spec)
        return version_spec

    def __post_init__(self):
        if self.use_this is None:
            if self.migration_guide is None:
                raise ValueError(
                    "@deprecate must include a migration guide if there is no replacement."
                )
        self.warn_starting = deprecated._ensure_version_type(self.warn_starting)
        self.fail_starting = deprecated._ensure_version_type(self.fail_starting)
        self.current_version = deprecated._ensure_version_type(self.current_version)
        self._validate_fail_starting()

    def _validate_fail_starting(self):
        if self.fail_starting.major > 0:  # This means we're past alpha. We are, but nice to have...
            if self.fail_starting.minor != 0 or self.fail_starting.patch != 0:
                raise ValueError(
                    f"Can only deprecate starting on major version releases. {self.fail_starting} is not valid."
                )
        if self.warn_starting > self.fail_starting:
            raise ValueError(
                f"warn_starting must come before fail_starting. {self.fail_starting} < {self.warn_starting}"
            )

    def _do_deprecation_action(self, fn: Callable):
        if self._should_fail():
            failure_message = " ".join(
                [
                    f"{fn.__qualname__} has been deprecated, as of hamilton version: {self.fail_starting}.",
                    f"{self.explanation}",
                ]
                + (
                    [f"Instead, you should be using: {self.use_this.__qualname__}."]
                    if self.use_this is not None
                    else []
                )
                + (
                    [f"For migration, see: {self.migration_guide}."]
                    if self.migration_guide is not None
                    else ["This is a drop-in replacement."]
                )
            )
            self.fail_action(failure_message)
        elif self._should_warn():
            warn_message = " ".join(
                [
                    f"{fn.__qualname__} will be deprecated by hamilton version: {self.fail_starting}.",
                    f"{self.explanation}",
                ]
                + (
                    [f"Instead, you should be using: {self.use_this.__qualname__}."]
                    if self.use_this is not None
                    else []
                )
                + (
                    [f"For migration, see: {self.migration_guide}."]
                    if self.migration_guide is not None
                    else ["This is a drop-in replacement."]
                )
            )
            self.warn_action(warn_message)

    def _should_warn(self) -> bool:
        return self.current_version > self.warn_starting

    def _should_fail(self) -> bool:
        return self.current_version > self.fail_starting

    def __call__(self, fn: Callable):
        """Decorates the function with the deprecated decorator.
        Note that this has different implementations for functions and
        objects that masquerade as functions (E.G. by implementing __call__.

        TODO -- use @singledispatchmethod  -- this was written before that was available
        https://docs.python.org/3/library/functools.html#functools.singledispatchmethod

        :param fn: function (or class) to decorate
        :return: The decorated function.
        """
        # In this case we just do a standard decorator
        if isinstance(fn, types.FunctionType):

            @functools.wraps(fn)
            def new_fn(*args, **kwargs):
                self._do_deprecation_action(fn)
                return fn(*args, **kwargs)

            return new_fn

        # Otherwise we assume that this is a class object
        # We have to store the former __call__
        fn.__old_call__ = fn.__call__

        def new__call__(self, *args, deprecator=self, fn=fn, **kwargs):
            deprecator._do_deprecation_action(fn)
            return fn.__old_call__(self, *args, **kwargs)

        # This works as we're assigning it to the entire class...
        # This also means we can't decorate it with 2 deprecation functions, but, hey,
        # its on you if you try to deprecate something twice
        fn.__call__ = new__call__
        return fn



from typing import List

import pandas as pd

from hamilton.function_modifiers import (
    UpstreamDependency,
    base,
    parameterize_extract_columns,
    source,
    value,
)
from hamilton.function_modifiers.expanders import ParameterizedExtract


def _get_dep_type(dep_type: str) -> UpstreamDependency:
    """Converts dependency type to the type known by function_modifier"""
    if dep_type == "out":
        return None
    if dep_type == "value":
        return value
    if dep_type == "source":
        return source
    raise ValueError(f"Invalid dep type: {dep_type}")


def _get_index_levels(index: pd.MultiIndex) -> List[list]:
    out = [[] for _ in index[0]]
    for specific_index in index:
        for i, key in enumerate(specific_index):
            out[i].append(key)
    return out


def _validate_df_parameterization(parameterization: pd.DataFrame):
    # TODO -- validate that its a multi-index
    columns = _get_index_levels(parameterization.columns)
    if (not len(columns) == 2) or "out" not in columns[1]:
        raise base.InvalidDecoratorException(
            "Decorator must have a double-index -- first index should be a "
            "list of {output, source, value} strs. Second must be a list of "
            "arguments in your function."
        )


def _convert_params_from_df(parameterization: pd.DataFrame) -> List[ParameterizedExtract]:
    _validate_df_parameterization(parameterization)
    args, dep_types = _get_index_levels(parameterization.columns)
    dep_types_converted = [_get_dep_type(val) for val in dep_types]
    out = []
    for _, column_set in parameterization.iterrows():
        parameterization = {
            arg: dep_type(col_value)
            for arg, col_value, dep_type in zip(args, column_set, dep_types_converted)
            if dep_type is not None
        }
        extracted_columns = [
            col for col, dep_type in zip(column_set, dep_types) if dep_type == "out"
        ]
        out.append(ParameterizedExtract(tuple(extracted_columns), parameterization))
    return out


class parameterize_frame(parameterize_extract_columns):
    """EXPERIMENTAL! Instantiates a parameterize_extract decorator using a dataframe to specify a set of extracts + \
    parameterizations.

    This is an experimental decorator and the API may change in the future; please provide feedback \
    whether this API does or does not work for you.

    :param parameterization: Parameterization dataframe. See below.

    This is of a specific shape:

        1. Index - Level 0: list of parameter names
        2. Index - Level 1: types of things to inject, either:

            - "out" (meaning this is an output),
            - "value" (meaning this is a literal value)
            - "source" (meaning this node comes from an upstream value)

        3. Contents:

          - Each row corresponds to the index. Each of these corresponds to an output node from this.


    Note your function has to take in the column-names and output a dataframe with those names -- \
    we will likely change it so that's not the case, and it can just use the position of the columns.

    Example usage:

    .. code-block:: python

        from hamilton.experimental.decorators.parameterize_frame import parameterize_frame
        df = pd.DataFrame(
        [
           ["outseries1a", "outseries2a", "inseries1a", "inseries2a", 5.0],
           ["outseries1b", "outseries2b", "inseries1b", "inseries2b", 0.2],
        ],
        # specify column names corresponding to function arguments and
        # if outputting multiple columns, output dataframe columns.
        columns=[
           ["output1", "output2", "input1", "input2", "input3"],
           ["out", "out", "source", "source", "value"],
        ])

        @parameterize_frame(df)
        def my_func(
            input1: pd.Series, input2: pd.Series, input3: float
        ) -> pd.DataFrame:
           ...

    """

    def __init__(self, parameterization: pd.DataFrame):
        super(parameterize_frame, self).__init__(*_convert_params_from_df(parameterization))


# Examples below
if __name__ == "__main__":
    df = pd.DataFrame(
        [
            ["outseries1a", "outseries2a", "inseries1a", "inseries2a", 5.0],
            ["outseries1b", "outseries2b", "inseries1b", "inseries2b", 0.2],
            # ...
        ],
        # Have to switch as indices have to be unique
        columns=[
            [
                "output1",
                "output2",
                "input1",
                "input2",
                "input3",
            ],
            # configure whether column is source or value and also whether it's input ("source", "value") or output ("out")
            ["out", "out", "source", "source", "value"],
        ],
    )  # specify column names (corresponding to function arguments and (if outputting multiple columns) output dataframe columns)

    @parameterize_frame(df)
    def my_func(input1: pd.Series, input2: pd.Series, input3: float) -> pd.DataFrame:
        return pd.DataFrame(
            [input1 * input2 * input3, input1 + input2 + input3]
        )  # if there's a single column it could maybe just return a series instead and pick up the name from the first column of the dataframe

    @parameterize_extract_columns(
        ParameterizedExtract(
            ("outseries1a", "outseries2a"),
            {"input1": source("inseries1a"), "input2": source("inseries2a"), "input3": value(5.0)},
        ),
        ParameterizedExtract(
            ("outseries1b", "outseries2b"),
            {"input1": source("inseries1b"), "input2": source("inseries2b"), "input3": value(0.2)},
        ),
    )
    def my_func_parameterized_extract(
        input1: pd.Series, input2: pd.Series, input3: float
    ) -> pd.DataFrame:
        print("running my_func_parameterized_extract")
        return pd.concat([input1 * input2 * input3, input1 + input2 + input3], axis=1)

    my_func_parameterized_extract.decorated = "false"

    # Test by running the @parameterized_extract decorator
    from hamilton.ad_hoc_utils import create_temporary_module
    from hamilton.driver import Driver

    dr = Driver({}, create_temporary_module(my_func_parameterized_extract))
    dr.visualize_execution(
        final_vars=["outseries1a", "outseries1b", "outseries2a", "outseries2b"],
        output_file_path="./out1.pdf",
        render_kwargs={},
        inputs={
            "inseries1a": pd.Series([1, 2]),
            "inseries1b": pd.Series([2, 3]),
            "inseries2a": pd.Series([3, 4]),
            "inseries2b": pd.Series([4, 5]),
        },
    )

    df_1 = dr.execute(
        final_vars=["outseries1a", "outseries1b", "outseries2a", "outseries2b"],
        # final_vars=["outseries1a", "outseries2a"],
        inputs={
            "inseries1a": pd.Series([1, 2]),
            "inseries1b": pd.Series([2, 3]),
            "inseries2a": pd.Series([3, 4]),
            "inseries2b": pd.Series([4, 5]),
        },
    )
    print(df_1)

    # Test by running the @parameterized_extract decorator
    dr = Driver({}, create_temporary_module(my_func))
    dr.visualize_execution(
        final_vars=["outseries1a", "outseries1b", "outseries2a", "outseries2b"],
        output_file_path="./out2.pdf",
        render_kwargs={},
        inputs={
            "inseries1a": pd.Series([1, 2]),
            "inseries1b": pd.Series([2, 3]),
            "inseries2a": pd.Series([3, 4]),
            "inseries2b": pd.Series([4, 5]),
        },
    )

    df_2 = dr.execute(
        final_vars=["outseries1a", "outseries1b", "outseries2a", "outseries2b"],
        # final_vars=["outseries1a", "outseries2a"],
        inputs={
            "inseries1a": pd.Series([1, 2]),
            "inseries1b": pd.Series([2, 3]),
            "inseries2a": pd.Series([3, 4]),
            "inseries2b": pd.Series([4, 5]),
        },
    )
    print(df_2)



import abc
import dataclasses
import functools
import logging
from concurrent.futures import Executor, Future, ProcessPoolExecutor, ThreadPoolExecutor
from typing import Any, Callable, Dict, List

from hamilton import node
from hamilton.execution.graph_functions import execute_subdag
from hamilton.execution.grouping import NodeGroupPurpose, TaskImplementation
from hamilton.execution.state import ExecutionState, GraphState, TaskState

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class TaskFuture:
    """Simple representation of a future. TODO -- add cancel().
    This a clean wrapper over a python future, and we may end up just using that at some point."""

    get_state: Callable[[], TaskState]
    get_result: Callable[[], Any]


class TaskExecutor(abc.ABC):
    """Abstract class for a task executor. All this does is submit a task and return a future.
    It also tells us if it can do that"""

    @abc.abstractmethod
    def init(self):
        """Initializes the task executor, provisioning any necessary resources."""
        pass

    @abc.abstractmethod
    def finalize(self):
        """Tears down the task executor, freeing up any provisioned resources.
        Will be called in a finally block."""
        pass

    @abc.abstractmethod
    def submit_task(self, task: TaskImplementation) -> TaskFuture:
        """Submits a task to the executor. Returns a task ID that can be used to query the status.
        Effectively a future.

        :param task: Task implementation (bound with arguments) to submit
        :return: The future representing the task's computation.
        """
        pass

    @abc.abstractmethod
    def can_submit_task(self) -> bool:
        """Returns whether or not we can submit a task to the executor.
        For instance, if the maximum parallelism is reached, we may not be able to submit a task.

        TODO -- consider if this should be a "parallelism" value instead of a boolean, forcing
        the ExecutionState to store the state prior to executing a task.

        :return: whether or not we can submit a task.
        """
        pass


def new_callable(*args, _callable=None, **kwargs):
    return list(_callable(*args, **kwargs))


def _modify_callable(node_source: node.NodeType, callabl: Callable):
    """This is a bit of a shortcut -- we modify the callable here as
    we want to allow `Parallelizable[]` nodes to return a generator

    :param node_source:
    :param callabl:
    :return:
    """
    if node_source == node.NodeType.EXPAND:
        return functools.partial(new_callable, _callable=callabl)
    return callabl


def base_execute_task(task: TaskImplementation) -> Dict[str, Any]:
    """This is a utility function to execute a base task. In an ideal world this would be recursive,
    (as in we can use the same task execution/management system as we would otherwise)
    but for now we just call out to good old DFS. Note that this only returns the result that
    a task is required to output, and does not return anything else. It also returns
    any overrides.

    We should probably have a simple way of doing this for single-node tasks, as they're
    going to be common.

    :param task: task to execute.
    :return: a diciontary of the results of all the nodes in that task's nodes to compute.
    """
    # We do this as an edge case to force the callable to return a list if it is an expand,
    # and would normally return a generator. That said, we will likely remove this in the future --
    # its an implementation detail, and a true queuing system between nodes/controller would mean
    # we wouldn't need to do this, and instead could just use the generator aspect.
    # Furthermore, in most cases the user wouldn't be calling an expand on a "remote" node,
    # but it is a supported use-case.
    for node_ in task.nodes:
        if not getattr(node_, "callable_modified", False):
            node_._callable = _modify_callable(node_.node_role, node_.callable)
        node_.callable_modified = True
    if task.adapter.does_hook("pre_task_execute", is_async=False):
        task.adapter.call_all_lifecycle_hooks_sync(
            "pre_task_execute",
            run_id=task.run_id,
            task_id=task.task_id,
            nodes=task.nodes,
            inputs=task.dynamic_inputs,
            overrides=task.overrides,
        )
    error = None
    success = True
    results = None
    try:
        results = execute_subdag(
            nodes=task.nodes,
            inputs=task.dynamic_inputs,
            adapter=task.adapter,
            overrides={**task.dynamic_inputs, **task.overrides},
            run_id=task.run_id,
            task_id=task.task_id,
        )
    except Exception as e:
        logger.exception(task.task_id)
        error = e
        success = False
        logger.exception(
            f"Exception executing task {task.task_id}, with nodes: {[item.name for item in task.nodes]}"
        )
        raise e
    finally:
        if task.adapter.does_hook("post_task_execute", is_async=False):
            task.adapter.call_all_lifecycle_hooks_sync(
                "post_task_execute",
                run_id=task.run_id,
                task_id=task.task_id,
                nodes=task.nodes,
                results=results,
                success=success,
                error=error,
            )
    # This selection is for GC
    # We also need to get the override values
    # This way if its overridden we can ensure it gets passed to the right one
    final_retval = {
        key: value
        for key, value in results.items()
        if key in task.outputs_to_compute or key in task.overrides
    }
    return final_retval


class SynchronousLocalTaskExecutor(TaskExecutor):
    """Basic synchronous/local task executor that runs tasks
    in the same process, at submit time."""

    def submit_task(self, task: TaskImplementation) -> TaskFuture:
        """Submitting a task is literally just running it.

        :param task: Task to submit
        :return: Future associated with this task
        """
        # No error management for now
        result = base_execute_task(task)
        return TaskFuture(get_state=lambda: TaskState.SUCCESSFUL, get_result=lambda: result)

    def can_submit_task(self) -> bool:
        """We can always submit a task as the task submission is blocking!

        :return: True
        """
        return True

    def init(self):
        pass

    def finalize(self):
        pass


class TaskFutureWrappingPythonFuture(TaskFuture):
    """Wraps a python future in a TaskFuture"""

    def __init__(self, future: Future):
        self.future = future

    def get_state(self):
        """Gets the state. This is non-blocking."""

        if self.future.done():
            try:
                self.future.result()
            except Exception:
                logger.exception("Task failed")
                return TaskState.FAILED
            return TaskState.SUCCESSFUL
        else:
            return TaskState.RUNNING

    def get_result(self):
        """Gets the result. This is non-blocking.

        :return: None if there is no result, else the result
        """
        if not self.future.done():
            return None
        out = self.future.result()
        return out


class PoolExecutor(TaskExecutor, abc.ABC):
    """Base class for a pool-based executor (threadpool executor/multiprocessing executor).
    Handles common logic, stores active future, and manages max tasks.
    """

    def __init__(self, max_tasks: int):
        self.active_futures = []
        self.initialized = False
        self.pool = None
        self.max_tasks = max_tasks  # TODO -- allow infinite/no max.

    def _prune_active_futures(self):
        self.active_futures = [f for f in self.active_futures if not f.done()]

    @abc.abstractmethod
    def create_pool(self) -> Executor:
        """Creates a pool to submit tasks to.

        :return: The executor to handle all the tasks in the pool
        """
        pass

    def init(self):
        """Creates/initializes pool"""
        if not self.initialized:
            self.pool = self.create_pool()
            self.initialized = True
        else:
            raise RuntimeError("Cannot initialize an already initialized executor")

    def finalize(self):
        """Finalizes pool, freeing up resources"""
        if self.initialized:
            self.pool.shutdown()
            self.initialized = False
        else:
            raise RuntimeError("Cannot finalize an uninitialized executor")

    def submit_task(self, task: TaskImplementation) -> TaskFuture:
        """Submitting a task is literally just running it.

        :param task: Task to submit
        :return: The future associated with the task
        """
        # First submit it
        # Then we need to wrap it in a future
        future = self.pool.submit(base_execute_task, task)
        self.active_futures.append(future)
        return TaskFutureWrappingPythonFuture(future)

    def can_submit_task(self) -> bool:
        """Tells if the pool is full or not.

        :return:
        """

        self._prune_active_futures()
        return len(self.active_futures) < self.max_tasks


class MultiThreadingExecutor(PoolExecutor):
    """Basic synchronous/local task executor that runs tasks
    in the same process, at submit time."""

    def create_pool(self) -> ThreadPoolExecutor:
        return ThreadPoolExecutor(max_workers=self.max_tasks)


class MultiProcessingExecutor(PoolExecutor):
    """Basic synchronous/local task executor that runs tasks
    in the same process, at submit time. Note that this is
    not yet augmented to handle the right serialization,
    so use at your own risk. We will be fixing shortly,
    but the dask/ray parallelism and the multithreading
    parallelism executors serialize correctly."""

    def create_pool(self) -> ProcessPoolExecutor:
        return ProcessPoolExecutor(max_workers=self.max_tasks)


class ExecutionManager(abc.ABC):
    """Manages execution per task. This enables you to have different executors for different
    tasks/task types. Note that, currently, it just uses the task information, but we could
    theoretically add metadata in a task as well.
    """

    def __init__(self, executors: List[TaskExecutor]):
        """Initializes the execution manager. Note this does not start it up/claim resources --
        you need to call init() to do that.

        :param executors:
        """
        self.executors = executors

    def init(self):
        """Initializes each of the executors."""
        for executor in self.executors:
            executor.init()

    def finalize(self):
        """Finalizes each of the executors."""
        for executor in self.executors:
            executor.finalize()

    @abc.abstractmethod
    def get_executor_for_task(self, task: TaskImplementation) -> TaskExecutor:
        """Selects the executor for the task. This enables us to set the appropriate executor
        for specific tasks (so that we can run some locally, some remotely, etc...).

        Note that this is the power-user case -- in all likelihood, people will use the default
        ExecutionManager.

        :param task:  Task to choose execution manager for
        :return: The executor to use for this task
        """
        pass


class DefaultExecutionManager(ExecutionManager):
    def __init__(self, local_executor=None, remote_executor=None):
        """Instantiates a BasicExecutionManager with a local/remote executor.
        These enable us to run certain tasks locally (simple transformations, generating sets of files),
        and certain tasks remotely (processing files in large datasets, etc...)

        :param local_executor: Executor to use for running tasks locally
        :param remote_executor:  Executor to use for running tasks remotely
        """
        if local_executor is None:
            local_executor = SynchronousLocalTaskExecutor()
        if remote_executor is None:
            remote_executor = MultiProcessingExecutor(max_tasks=5)
        super().__init__([local_executor, remote_executor])
        self.local_executor = local_executor
        self.remote_executor = remote_executor

    def get_executor_for_task(self, task: TaskImplementation) -> TaskExecutor:
        """Simple implementation that returns the local executor for single task executions,

        :param task: Task to get executor for
        :return: A local task if this is a "single-node" task, a remote task otherwise
        """

        if task.purpose == NodeGroupPurpose.EXECUTE_BLOCK:
            return self.remote_executor
        return self.local_executor


def run_graph_to_completion(
    execution_state: ExecutionState,
    execution_manager: ExecutionManager,
):
    """Blocking call to run the graph until it is complete. Note that this employs a while loop.
    With the way we handle futures, we should be able to have this event-driven, allowing us
    to only query when the state is updated, and use that to trigger a new update.

    For now, the while loop is fine.

    :return: Nothing, the execution state/result cache can give us the data
    """
    task_futures = {}
    execution_manager.init()
    try:
        while not GraphState.is_terminal(execution_state.get_graph_state()):
            # get the next task from the queue
            next_task = execution_state.release_next_task()
            if next_task is not None:
                task_executor = execution_manager.get_executor_for_task(next_task)
                if task_executor.can_submit_task():
                    try:
                        submitted = task_executor.submit_task(next_task)
                    except Exception as e:
                        logger.exception(
                            f"Exception submitting task {next_task.task_id}, with nodes: "
                            f"{[item.name for item in next_task.nodes]}"
                        )
                        raise e
                    task_futures[next_task.task_id] = submitted
                else:
                    # Whoops, back on the queue
                    # We should probably wait a bit here, but for now we're going to keep
                    # burning through
                    execution_state.reject_task(task_to_reject=next_task)
            # update all the tasks in flight
            # copy so we can modify
            for task_name, task_future in task_futures.copy().items():
                state = task_future.get_state()
                result = task_future.get_result()
                execution_state.update_task_state(task_name, state, result)
                if TaskState.is_terminal(state):
                    del task_futures[task_name]
        logger.info(f"Graph is done, graph state is {execution_state.get_graph_state()}")
    finally:
        execution_manager.finalize()



import base64
import zlib
from typing import Dict, Mapping


def _compress_string(string: str) -> str:
    return base64.b64encode(zlib.compress(string.encode(), level=3)).decode()


def _decompress_string(string: str) -> str:
    return zlib.decompress(base64.b64decode(string.encode())).decode()


def _encode_str_dict(d: Mapping) -> str:
    interleaved_tuple = tuple(item for pair in sorted(d.items()) for item in pair)
    return ",".join(interleaved_tuple)


def _decode_str_dict(s: str) -> Mapping:
    interleaved_tuple = tuple(s.split(","))
    d = {}
    for i in range(0, len(interleaved_tuple), 2):
        d[interleaved_tuple[i]] = interleaved_tuple[i + 1]
    return d


def decode_key(cache_key: str) -> dict:
    node_name, _, code_and_data_string = cache_key.partition("-")
    code_version, _, dep_encoded = code_and_data_string.partition("-")
    data_stringified = _decompress_string(dep_encoded)

    if data_stringified == "<none>":
        dependencies_data_versions = {}
    else:
        dependencies_data_versions = _decode_str_dict(data_stringified)
    return dict(
        node_name=node_name,
        code_version=code_version,
        dependencies_data_versions=dependencies_data_versions,
    )


def create_cache_key(
    node_name: str, code_version: str, dependencies_data_versions: Dict[str, str]
) -> str:
    if len(dependencies_data_versions.keys()) > 0:
        dependencies_stringified = _encode_str_dict(dependencies_data_versions)
    else:
        dependencies_stringified = "<none>"

    safe_node_name = "".join(c for c in node_name if c.isalnum() or c in ("_",)).rstrip()

    return f"{safe_node_name}-{code_version}-{_compress_string(dependencies_stringified)}"



import collections
import configparser
import functools
import importlib
import logging
import os
import pathlib
from typing import Any, Dict, Literal, Optional, Tuple, Type, get_args

logger = logging.getLogger(__name__)

# Use this to ensure the registry is loaded only once.
INITIALIZED = False
ExtensionName = Literal[
    "yaml",
    "matplotlib",
    "numpy",
    "pandas",
    "plotly",
    "polars",
    "polars_lazyframe",
    "pyspark_pandas",
    "spark",
    "dask",
    "geopandas",
    "xgboost",
    "lightgbm",
    "sklearn_plot",
    "vaex",
    "ibis",
    "dlt",
    "kedro",
    "huggingface",
    "mlflow",
    "pydantic",
]
HAMILTON_EXTENSIONS: Tuple[ExtensionName, ...] = get_args(ExtensionName)
HAMILTON_AUTOLOAD_ENV = "HAMILTON_AUTOLOAD_EXTENSIONS"
# NOTE the variable DEFAULT_CONFIG_LOCAITON is redundant with `hamilton.telemetry`
# but this `registry` module must avoid circular imports
DEFAULT_CONFIG_LOCATION = pathlib.Path("~/.hamilton.conf").expanduser()

# This is a dictionary of extension name -> dict with dataframe and column types.
DF_TYPE_AND_COLUMN_TYPES: Dict[str, Dict[str, Type]] = {}

COLUMN_TYPE = "column_type"
DATAFRAME_TYPE = "dataframe_type"


def load_autoload_config() -> configparser.ConfigParser:
    """Load the Hamilton config file and set the autoloading environment variable"""
    config = configparser.ConfigParser()
    config.read(DEFAULT_CONFIG_LOCATION)

    if config.has_option("DEFAULT", HAMILTON_AUTOLOAD_ENV):
        os.environ[HAMILTON_AUTOLOAD_ENV] = config.get("DEFAULT", HAMILTON_AUTOLOAD_ENV)

    return config


load_autoload_config()


def load_extension(plugin_module: ExtensionName):
    """Given a module name, loads it for Hamilton to use.

    :param plugin_module: the module name sans .py. e.g. pandas, polars, pyspark_pandas.
    """
    mod = importlib.import_module(f"hamilton.plugins.{plugin_module}_extensions")
    # We have various plugin extensions. We default to assuming it's a dataframe extension with columns,
    # unless it explicitly says it's not.
    # We need to check the following if we are to enable `@extract_columns` for example.
    extractable = getattr(mod, "COLUMN_FRIENDLY_DF_TYPE", True)
    if extractable:
        assert hasattr(mod, "register_types"), "Error extension missing function register_types()"
        assert hasattr(
            mod, f"get_column_{plugin_module}"
        ), f"Error extension missing get_column_{plugin_module}"
        assert hasattr(
            mod, f"fill_with_scalar_{plugin_module}"
        ), f"Error extension missing fill_with_scalar_{plugin_module}"
        logger.info(f"Detected {plugin_module} and successfully loaded Hamilton extensions.")


def initialize():
    """Iterate over all extensions and try to load them"""
    logger.debug(f"{HAMILTON_AUTOLOAD_ENV}={os.environ.get(HAMILTON_AUTOLOAD_ENV)}")
    for extension_name in HAMILTON_EXTENSIONS:
        # skip modules that aren't explicitly imported by the user
        if str(os.environ.get(HAMILTON_AUTOLOAD_ENV)) == "0":
            continue

        try:
            load_extension(extension_name)
        except NotImplementedError as e:
            logger.debug(f"Did not load {extension_name} extension because {str(e)}.")
        except ModuleNotFoundError as e:
            logger.debug(f"Did not load {extension_name} extension because {e.msg}.")
        except ImportError as e:
            logger.debug(f"Did not load {extension_name} extension because {str(e)}.")

    global INITIALIZED
    INITIALIZED = True


def disable_autoload():
    """Disable extension autoloading by setting an environment variable.
    This needs to be done before hamilton.driver is imported.
    """
    os.environ[HAMILTON_AUTOLOAD_ENV] = "0"


def enable_autoload():
    """Enable extension autoloading by deleting an environment variable.
    This needs to be done before hamilton.driver is imported.
    """
    del os.environ[HAMILTON_AUTOLOAD_ENV]


def config_enable_autoload():
    """Modify the Hamilton config file to enable extension autoloading.
    Autoloading can be disabled manually via `hamilton.registry.disable_autoload()`
    before importing `hamilton.driver`.

    NOTE the function name is tied to an entrypoint in `pyproject.toml`
    """
    config = load_autoload_config()
    if "DEFAULT" not in config:
        config.add_section("DEFAULT")

    config.remove_option("DEFAULT", HAMILTON_AUTOLOAD_ENV)
    with DEFAULT_CONFIG_LOCATION.open("w") as f:
        config.write(f)


def config_disable_autoload():
    """Modify the Hamilton config file to disable extension autoloading.
    Autoloading can be enabled manually via `hamilton.registry.enable_autoload()`
    before importing `hamilton.driver`.

    NOTE the function name is tied to an entrypoint in `pyproject.toml`
    """
    config = load_autoload_config()
    if "DEFAULT" not in config:
        config.add_section("DEFAULT")

    config.set("DEFAULT", HAMILTON_AUTOLOAD_ENV, "0")
    with DEFAULT_CONFIG_LOCATION.open("w") as f:
        config.write(f)


def register_types(extension_name: str, dataframe_type: Type, column_type: Optional[Type]):
    """Registers the dataframe and column types for the extension. Note that column types are optional
    as some extensions may not have a column type (E.G. spark). In this case, this is not included

    :param extension_name: name of the extension doing the registering.
    :param dataframe_type: the dataframe type to register.
    :param column_type: the column type to register
    """
    global DF_TYPE_AND_COLUMN_TYPES
    output = {}
    output[DATAFRAME_TYPE] = dataframe_type
    if column_type is not None:
        output[COLUMN_TYPE] = column_type
    DF_TYPE_AND_COLUMN_TYPES[extension_name] = output


@functools.singledispatch
def get_column(df: Any, column_name: str):
    """Gets a column from a dataframe.

    Each extension should register a function for this.

    :param df: the dataframe.
    :param column_name: the column name.
    :return: the correct "representation" of a column for this "dataframe".
    """
    raise NotImplementedError()


@functools.singledispatch
def fill_with_scalar(df: Any, column_name: str, scalar_value: Any) -> Any:
    """Fills a column with a scalar value.

    :param df: the dataframe.
    :param column_name: the column to fill.
    :param scalar_value: the scalar value to fill with.
    :return: the modified dataframe.
    """
    raise NotImplementedError()


def get_column_type_from_df_type(dataframe_type: Type) -> Type:
    """Function to cycle through the registered extensions and return the column type for the dataframe type.

    :param dataframe_type: the dataframe type to find the column type for.
    :return: the column type.
    :raises: NotImplementedError if we don't know what the column type is.
    """
    for _extension, type_map in DF_TYPE_AND_COLUMN_TYPES.items():
        if dataframe_type == type_map[DATAFRAME_TYPE]:
            return type_map[COLUMN_TYPE]
    raise NotImplementedError(
        f"Cannot get column type for [{dataframe_type}]. "
        f"Registered types are {DF_TYPE_AND_COLUMN_TYPES}"
    )


LOADER_REGISTRY = collections.defaultdict(list)
SAVER_REGISTRY = collections.defaultdict(list)


def register_adapter(adapter: Any):
    """Registers a adapter. Note that the type is any,
    because we can't import it here due to circular imports.

    :param adapter: the adapter to register.
    """
    if adapter.can_load():
        LOADER_REGISTRY[adapter.name()].append(adapter)
    if adapter.can_save():
        SAVER_REGISTRY[adapter.name()].append(adapter)


def get_registered_dataframe_types() -> Dict[str, Type]:
    """Returns a dictionary of extension name -> dataframe type.

    :return: the dictionary.
    """
    return {
        extension: type_map[DATAFRAME_TYPE]
        for extension, type_map in DF_TYPE_AND_COLUMN_TYPES.items()
    }


def get_registered_column_types() -> Dict[str, Type]:
    """Returns a dictionary of extension name -> column type.

    :return: the dictionary.
    """
    return {
        extension: type_map[COLUMN_TYPE] for extension, type_map in DF_TYPE_AND_COLUMN_TYPES.items()
    }



from hamilton.htypes import Collect, Parallelizable


def number_of_steps() -> int:
    return 5


def param_external_to_block() -> int:
    return 3


def second_param_external_to_block() -> int:
    return 4


def steps(number_of_steps: int) -> Parallelizable[int]:
    yield from range(number_of_steps)


# Parallelizable block Start


def step_modified(steps: int, second_param_external_to_block: int) -> int:
    return steps + second_param_external_to_block


def double_step(step_modified: int) -> int:
    return step_modified * 2


def triple_step(step_modified: int) -> int:
    return step_modified * 3


def double_plus_triple_step(double_step: int, triple_step: int) -> int:
    return double_step + triple_step


def double_plus_triple_plus_param_external_to_block(
    double_plus_triple_step: int, param_external_to_block: int
) -> int:
    return double_plus_triple_step + param_external_to_block


# Parallelizable block ends here


def sum_of_some_things(double_plus_triple_plus_param_external_to_block: Collect[int]) -> int:
    return sum(double_plus_triple_plus_param_external_to_block)


def final(sum_of_some_things: int) -> int:
    return sum_of_some_things



import abc
import functools
import json
import logging
import time
import typing

import ray
from ray import workflow

from hamilton import base, htypes, lifecycle, node
from hamilton.execution import executors
from hamilton.execution.executors import TaskFuture
from hamilton.execution.grouping import TaskImplementation
from hamilton.function_modifiers.metadata import RAY_REMOTE_TAG_NAMESPACE

logger = logging.getLogger(__name__)


def raify(fn):
    """Makes the function into something ray-friendly.
    This is necessary due to https://github.com/ray-project/ray/issues/28146.

    :param fn: Function to make ray-friendly
    :return: The ray-friendly version
    """
    if isinstance(fn, functools.partial):

        def new_fn(*args, **kwargs):
            return fn(*args, **kwargs)

        return new_fn
    return fn


def parse_ray_remote_options_from_tags(tags: typing.Dict[str, str]) -> typing.Dict[str, typing.Any]:
    """DRY helper to parse ray.remote(**options) from Hamilton Tags

    Tags are added to nodes via the @ray_remote_options decorator

    :param tags: Full set of Tags for a Node
    :return: The ray-friendly version
    """

    ray_tags = {
        tag_name: tag_value
        for tag_name, tag_value in tags.items()
        if tag_name.startswith(f"{RAY_REMOTE_TAG_NAMESPACE}.")
    }
    ray_options = {name.split(".", 1)[1]: json.loads(value) for name, value in ray_tags.items()}

    return ray_options


class RayGraphAdapter(
    lifecycle.base.BaseDoRemoteExecute,
    lifecycle.base.BaseDoBuildResult,
    lifecycle.base.BaseDoValidateInput,
    lifecycle.base.BaseDoCheckEdgeTypesMatch,
    lifecycle.base.BasePostGraphExecute,
    abc.ABC,
):
    """Class representing what's required to make Hamilton run on Ray.

    This walks the graph and translates it to run onto `Ray <https://ray.io/>`__.

    Use `pip install sf-hamilton[ray]` to get the dependencies required to run this.

    Use this if:

      * you want to utilize multiple cores on a single machine, or you want to scale to larger data set sizes with\
        a Ray cluster that you can connect to. Note (1): you are still constrained by machine memory size with Ray; you\
        can't just scale to any dataset size. Note (2): serialization costs can outweigh the benefits of parallelism \
        so you should benchmark your code to see if it's worth it.

    Notes on scaling:
    -----------------
      - Multi-core on single machine ✅
      - Distributed computation on a Ray cluster ✅
      - Scales to any size of data ⛔️; you are LIMITED by the memory on the instance/computer 💻.

    Function return object types supported:
    ---------------------------------------
      - Works for any python object that can be serialized by the Ray framework. ✅

    Pandas?
    --------
      - ⛔️ Ray DOES NOT do anything special about Pandas.

    CAVEATS
    -------
      - Serialization costs can outweigh the benefits of parallelism, so you should benchmark your code to see if it's\
      worth it.

    DISCLAIMER -- this class is experimental, so signature changes are a possibility!
    """

    def __init__(
        self,
        result_builder: base.ResultMixin,
        ray_init_config: typing.Dict[str, typing.Any] = None,
        shutdown_ray_on_completion: bool = False,
    ):
        """Constructor

        You have the ability to pass in a ResultMixin object to the constructor to control the return type that gets \
        produce by running on Ray.

        :param result_builder: Required. An implementation of base.ResultMixin.
        :param ray_init_config: allows to connect to an existing cluster or start a new one with custom configuration (https://docs.ray.io/en/latest/ray-core/api/doc/ray.init.html)
        :param shutdown_ray_on_completion: by default we leave the cluster open, but we can also shut it down

        """
        self.result_builder = result_builder
        if not self.result_builder:
            raise ValueError(
                "Error: ResultMixin object required. Please pass one in for `result_builder`."
            )

        self.shutdown_ray_on_completion = shutdown_ray_on_completion

        if ray_init_config is not None:
            ray.init(**ray_init_config)

    @staticmethod
    def do_validate_input(node_type: typing.Type, input_value: typing.Any) -> bool:
        # NOTE: the type of a raylet is unknown until they are computed
        if isinstance(input_value, ray._raylet.ObjectRef):
            return True
        return htypes.check_input_type(node_type, input_value)

    @staticmethod
    def do_check_edge_types_match(type_from: typing.Type, type_to: typing.Type) -> bool:
        return type_from == type_to

    def do_remote_execute(
        self,
        *,
        execute_lifecycle_for_node: typing.Callable,
        node: node.Node,
        **kwargs: typing.Dict[str, typing.Any],
    ) -> typing.Any:
        """Function that is called as we walk the graph to determine how to execute a hamilton function.

        :param execute_lifecycle_for_node: wrapper function that executes lifecycle hooks and methods
        :param kwargs: the arguments that should be passed to it.
        :return: returns a ray object reference.
        """
        ray_options = parse_ray_remote_options_from_tags(node.tags)
        return ray.remote(raify(execute_lifecycle_for_node)).options(**ray_options).remote(**kwargs)

    def do_build_result(self, outputs: typing.Dict[str, typing.Any]) -> typing.Any:
        """Builds the result and brings it back to this running process.

        :param outputs: the dictionary of key -> Union[ray object reference | value]
        :return: The type of object returned by self.result_builder.
        """
        if logger.isEnabledFor(logging.DEBUG):
            for k, v in outputs.items():
                logger.debug(f"Got output {k}, with type [{type(v)}].")
        # need to wrap our result builder in a remote call and then pass in what we want to build from.
        remote_combine = ray.remote(self.result_builder.build_result).remote(**outputs)
        result = ray.get(remote_combine)  # this materializes the object locally
        return result

    def post_graph_execute(self, *args, **kwargs):
        """We have the option to close the cluster down after execution."""

        if self.shutdown_ray_on_completion:
            # In case we have Hamilton Tracker to have enough time to properly flush
            time.sleep(5)
            ray.shutdown()


class RayWorkflowGraphAdapter(base.HamiltonGraphAdapter, base.ResultMixin):
    """Class representing what's required to make Hamilton run Ray Workflows

    Use `pip install sf-hamilton[ray]` to get the dependencies required to run this.

    Ray workflows is a more robust way to scale computation for any type of Hamilton graph.

    What's the difference between this and RayGraphAdapter?
    --------------------------------------------------------
        * Ray workflows offer durable computation. That is, they save and checkpoint each function.
        * This enables one to run a workflow, and not have to restart it if something fails, assuming correct\
        Ray workflow usage.

    Tips
    ----
    See https://docs.ray.io/en/latest/workflows/basics.html for the source of the following:

        1. Functions should be idempotent.
        2. The workflow ID is what Ray uses to try to resume/restart if run a second time.
        3. Nothing is run until the entire DAG is walked and setup and build_result is called.

    Notes on scaling:
    -----------------
      - Multi-core on single machine ✅
      - Distributed computation on a Ray cluster ✅
      - Scales to any size of data ⛔️; you are LIMITED by the memory on the instance/computer 💻.

    Function return object types supported:
    ---------------------------------------
      - Works for any python object that can be serialized by the Ray framework. ✅

    Pandas?
    --------
      - ⛔️ Ray DOES NOT do anything special about Pandas.

    CAVEATS
    -------
      - Serialization costs can outweigh the benefits of parallelism, so you should benchmark your code to see if it's\
      worth it.

    DISCLAIMER -- this class is experimental, so signature changes are a possibility!
    """

    def __init__(self, result_builder: base.ResultMixin, workflow_id: str):
        """Constructor

        :param result_builder: Required. An implementation of base.ResultMixin.
        :param workflow_id: Required. An ID to give the ray workflow to identify it for durability purposes.
        :param max_retries: Optional. The function will be retried for the given number of times if an
            exception is raised.
        """
        self.result_builder = result_builder
        self.workflow_id = workflow_id
        if not self.result_builder:
            raise ValueError(
                "Error: ResultMixin object required. Please pass one in for `result_builder`."
            )

    @staticmethod
    def check_input_type(node_type: typing.Type, input_value: typing.Any) -> bool:
        # NOTE: the type of a raylet is unknown until they are computed
        if isinstance(input_value, ray._raylet.ObjectRef):
            return True
        return htypes.check_input_type(node_type, input_value)

    @staticmethod
    def check_node_type_equivalence(node_type: typing.Type, input_type: typing.Type) -> bool:
        return node_type == input_type

    def execute_node(self, node: node.Node, kwargs: typing.Dict[str, typing.Any]) -> typing.Any:
        """Function that is called as we walk the graph to determine how to execute a hamilton function.

        :param node: the node from the graph.
        :param kwargs: the arguments that should be passed to it.
        :return: returns a ray object reference.
        """
        ray_options = parse_ray_remote_options_from_tags(node.tags)
        return ray.remote(raify(node.callable)).options(**ray_options).bind(**kwargs)

    def build_result(self, **outputs: typing.Dict[str, typing.Any]) -> typing.Any:
        """Builds the result and brings it back to this running process.

        :param outputs: the dictionary of key -> Union[ray object reference | value]
        :return: The type of object returned by self.result_builder.
        """
        if logger.isEnabledFor(logging.DEBUG):
            for k, v in outputs.items():
                logger.debug(f"Got output {k}, with type [{type(v)}].")
        # need to wrap our result builder in a remote call and then pass in what we want to build from.
        remote_combine = ray.remote(self.result_builder.build_result).bind(**outputs)
        result = workflow.run(
            remote_combine, workflow_id=self.workflow_id
        )  # this materializes the object locally
        return result


class RayTaskExecutor(executors.TaskExecutor):
    """Task executor using Ray for the new task-based execution mechanism in Hamilton.
    This is still experimental, so the API might change.
    """

    def __init__(
        self,
        num_cpus: int = None,
        ray_init_config: typing.Dict[str, typing.Any] = None,
        skip_init: bool = False,
    ):
        """Creates a ray task executor. Note this will likely take in more parameters. This is
        experimental, so the API will likely change, although we will do our best to make it
        backwards compatible.

        :param num_cpus: Number of cores to use for initialization, passed directly to ray.init. Defaults to all cores.
        :param ray_init_config: General configuration to pass to ray.init. Defaults to None.
        :param skip_init: Skips ray init if you already have Ray initialized. Default is False.
        """
        self.num_cpus = num_cpus
        self.ray_init_config = ray_init_config if ray_init_config else {}
        self.skip_init = skip_init

    def init(self):
        if self.skip_init:
            return
        ray.init(num_cpus=self.num_cpus, **self.ray_init_config)

    def finalize(self):
        if self.skip_init:
            # we assume that if we didn't init it, we don't need to shutdown either.
            return
        ray.shutdown()

    def submit_task(self, task: TaskImplementation) -> TaskFuture:
        """Submits a task, wrapping it in a TaskFuture (after getting the corresponding python
        future).

        :param task: Task to wrap
        :return: A future
        """

        return executors.TaskFutureWrappingPythonFuture(
            ray.remote(executors.base_execute_task).remote(task=task).future()
        )

    def can_submit_task(self) -> bool:
        """For now we can always submit a task -- it might just be delayed.

        :return: True
        """
        return True



from typing import List

from pydantic import BaseModel

from hamilton import node
from hamilton.data_quality import base as dq_base
from hamilton.function_modifiers import InvalidDecoratorException
from hamilton.function_modifiers import base as fm_base
from hamilton.function_modifiers import check_output as base_check_output
from hamilton.function_modifiers.validation import BaseDataValidationDecorator
from hamilton.htypes import custom_subclass_check


class check_output(BaseDataValidationDecorator):
    def __init__(
        self,
        importance: str = dq_base.DataValidationLevel.WARN.value,
        target: fm_base.TargetType = None,
    ):
        """Specific output-checker for pydantic models (requires ``pydantic>=2.0``).
        This decorator utilizes the output type of the function, which can be any subclass of pydantic.BaseModel.
        The function output must be declared with a type hint.

        :param model: The pydantic model to use for validation. If this is not provided, then the output type of the function is used.
        :param importance: Importance level (either "warn" or "fail") -- see documentation for check_output for more details.
        :param target: The target of the decorator -- see documentation for check_output for more details.

        Here is an example of how to use this decorator with a function that returns a pydantic model:

        .. code-block:: python

            from hamilton.plugins import h_pydantic
            from pydantic import BaseModel

            class MyModel(BaseModel):
                a: int
                b: float
                c: str

            @h_pydantic.check_output()
            def foo() -> MyModel:
                return MyModel(a=1, b=2.0, c="hello")

        Alternatively, you can return a dictionary from the function (type checkers will probably
        complain about this):

        .. code-block:: python

            from hamilton.plugins import h_pydantic
            from pydantic import BaseModel

            class MyModel(BaseModel):
                a: int
                b: float
                c: str

            @h_pydantic.check_output()
            def foo() -> MyModel:
                return {"a": 1, "b": 2.0, "c": "hello"}

        You can also use pydantic validation through ``function_modifiers.check_output`` by
        providing the model as an argument:

        .. code-block:: python

            from typing import Any

            from hamilton import function_modifiers
            from pydantic import BaseModel

            class MyModel(BaseModel):
                a: int
                b: float
                c: str

            @function_modifiers.check_output(model=MyModel)
            def foo() -> dict[str, Any]:
                return {"a": 1, "b": 2.0, "c": "hello"}

        Note, that because we do not (yet) support modification of the output, the validation is
        performed in strict mode, meaning that no data coercion is performed. For example, the
        following function will *fail* validation:

        .. code-block:: python

            from hamilton.plugins import h_pydantic
            from pydantic import BaseModel

            class MyModel(BaseModel):
                a: int  # Defined as an int

            @h_pydantic.check_output()  # This will fail validation!
            def foo() -> MyModel:
                return MyModel(a="1")  # Assigned as a string

        For more information about strict mode see the pydantic docs: https://docs.pydantic.dev/latest/concepts/strict_mode/

        """
        super(check_output, self).__init__(target)
        self.importance = importance
        self.target = target

    def get_validators(self, node_to_validate: node.Node) -> List[dq_base.DataValidator]:
        output_type = node_to_validate.type
        if not custom_subclass_check(output_type, BaseModel):
            raise InvalidDecoratorException(
                f"Output of function {node_to_validate.name} must be a Pydantic model"
            )
        return base_check_output(
            importance=self.importance, model=output_type, target_=self.target
        ).get_validators(node_to_validate)



import json
import logging
from contextvars import ContextVar
from typing import Any, Collection, Dict, List, Optional, Tuple

logger = logging.getLogger(__name__)


try:
    from opentelemetry import context, trace
    from opentelemetry.sdk.trace import Span
except ImportError as e:
    raise ImportError(
        "Failed to import `opentelemetry` "
        "Use `pip install sf-hamilton[opentelemetry]` to install "
        "dependencies for the `h_opentelemetry` plugin."
    ) from e

from hamilton.graph_types import HamiltonGraph, HamiltonNode
from hamilton.lifecycle import GraphExecutionHook, NodeExecutionHook, TaskExecutionHook

# We have to keep track of tokens for the span
# As OpenTel has some weird behavior around context managers, we have to account for the latest ones we started
# This way we can pop one off and know where to set the current one (as the parent, when the next one ends)
token_stack = ContextVar[Optional[List[Tuple[object, Span]]]]("token_stack", default=None)


def _exit_span(exc: Optional[Exception] = None):
    """Ditto with _enter_span, but for exiting the span. Pops the token off the stack and detaches the context."""
    stack = token_stack.get()[:]
    token, span = stack.pop()
    token_stack.set(stack)
    context.detach(token)
    if exc:
        span.set_status(trace.Status(trace.StatusCode.ERROR, str(exc)))
    else:
        span.set_status(trace.Status(trace.StatusCode.OK))
    span.end()
    return span


def _enter_span(name: str, tracer: trace.Tracer):
    """Utility function to enter a span. Starts, sets the current context, and adds it to the token stack.

    See this for some background on why start_span doesn't really work. We could use start_as_current_span,
    but this is a bit more explicit.
    """
    span = tracer.start_span(
        name=name,
        record_exception=False,  # we'll handle this ourselves
        set_status_on_exception=False,
    )
    ctx = trace.set_span_in_context(span)
    token = context.attach(ctx)
    stack = (token_stack.get() or [])[:]
    stack.append((token, span))
    token_stack.set(stack)
    return span


class OpenTelemetryTracer(NodeExecutionHook, GraphExecutionHook, TaskExecutionHook):
    """Adapter to log Hamilton execution to OpenTelemetry. At a high level, this works as follows:
    1. On any of the start/pre hooks (run_before_graph, run_before_node, run_before_task), we start a new span
    2. On any of the post ones we exit the span, accounting for the error (setting it if needed)

    This works by logging to OpenTelemetry, and setting the span processor to be the right one (that knows about the tracker).
    """

    def __init__(self, tracer_name: Optional[str] = None, tracer: Optional[trace.Tracer] = None):
        if tracer_name and tracer:
            raise ValueError(
                f"Only pass in one of tracer_name or tracer, not both, got: tracer_name={tracer_name} and tracer={tracer}"
            )

        if tracer:
            self.tracer = tracer
        elif tracer_name:
            self.tracer = trace.get_tracer(tracer_name)
        else:
            self.tracer = trace.get_tracer(__name__)

        self.graph = None

    def run_before_graph_execution(
        self,
        *,
        graph: HamiltonGraph,
        final_vars: List[str],
        inputs: dict,
        overrides: dict,
        execution_path: Collection[str],
        run_id: str,
        **kwargs,
    ):
        self.graph = graph

        attributes = {
            "graph_version": graph.version,
            "final_vars": final_vars,
            "inputs": list(inputs.keys()) if inputs else [],
            "overrides": list(overrides.keys()) if overrides else [],
            "execution_path": list(execution_path),
        }

        graph_span = _enter_span(run_id, self.tracer)
        graph_span.set_attributes(attributes)

    def run_before_node_execution(
        self,
        *,
        node_name: str,
        node_tags: dict,
        node_return_type: type,
        **kwargs: Any,
    ):
        attributes = {
            "type": str(node_return_type),
            "node_version": self.graph[node_name].version,
            "tags": json.dumps(node_tags),
        }

        node_span = _enter_span(node_name, self.tracer)
        node_span.set_attributes(attributes)

    def run_before_task_execution(
        self,
        *,
        task_id: str,
        nodes: List[HamiltonNode],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
        **kwargs,
    ):
        attributes = {
            "nodes": [n.name for n in nodes],
            "inputs": list(inputs.keys()) if inputs else [],
            "overrides": list(overrides.keys()) if overrides else [],
        }
        task_span = _enter_span(task_id, self.tracer)
        task_span.set_attributes(attributes)

    def run_after_task_execution(self, *, error: Optional[Exception], **kwargs):
        _exit_span(error)

    def run_after_node_execution(self, *, error: Optional[Exception], **kwargs):
        _exit_span(error)

    def run_after_graph_execution(self, *, error: Optional[Exception], **kwargs):
        _exit_span(error)



import json
import sys
import traceback
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

import attr
from openlineage.client import OpenLineageClient, event_v2, facet_v2

from hamilton import graph as h_graph
from hamilton import graph_types, node
from hamilton.lifecycle import base


@attr.s
class HamiltonFacet(facet_v2.RunFacet):
    """Class for Hamilton Facet."""

    hamilton_run_id: str = attr.ib()
    graph_version: str = attr.ib()
    final_vars: List[str] = attr.ib()
    inputs: List[str] = attr.ib()
    overrides: List[str] = attr.ib()


def get_stack_trace(exception):
    # Python changed this API in 3.10
    if sys.version_info < (3, 10, 0):
        return traceback.format_exception(
            etype=type(exception), value=exception, tb=exception.__traceback__
        )

    return "".join(traceback.format_exception(exception))


def extract_schema_facet(metadata):
    """Extracts the schema facet from the metadata."""
    if "dataframe_metadata" in metadata:
        schema_datatypes = [
            facet_v2.schema_dataset.SchemaDatasetFacetFields(
                name=k,
                type=v,
            )
            for k, v in zip(
                metadata["dataframe_metadata"]["column_names"],
                metadata["dataframe_metadata"]["datatypes"],
            )
        ]
        schema_facet = facet_v2.schema_dataset.SchemaDatasetFacet(
            fields=schema_datatypes,
        )
        return schema_facet
    return None


def create_input_dataset(namespace: str, metadata: dict, node_) -> list[event_v2.InputDataset]:
    """Creates the open lineage input dataset."""
    datasource_facet = None
    storage_facet = None
    sql_facet = None
    if "file_metadata" in metadata:
        name = node_.name
        if ".loader" in name:
            name = name.split(".loader")[0]
        path = metadata["file_metadata"]["path"]
        format = path.split(".")[-1] if "." in name else "unknown"
        storage_facet = facet_v2.storage_dataset.StorageDatasetFacet(
            storageLayer="FileSystem",
            fileFormat=format,
        )
        datasource_facet = facet_v2.datasource_dataset.DatasourceDatasetFacet(
            name=name,
            uri=path,
        )
    elif "sql_metadata" in metadata:
        name = metadata["sql_metadata"]["table_name"]
        sql_facet = facet_v2.sql_job.SQLJobFacet(
            query=metadata["sql_metadata"]["query"],
        )
    else:
        name = "--UNKNOWN--"
    schema_facet = extract_schema_facet(metadata)
    inputFacets = {}
    if storage_facet:
        inputFacets["storage"] = storage_facet
    if datasource_facet:
        inputFacets["dataSource"] = datasource_facet
    if schema_facet:
        inputFacets["schema"] = schema_facet
    if len(inputFacets) == 0:
        inputFacets = None
    inputs = [event_v2.InputDataset(namespace, name, facets=inputFacets)]
    return inputs, sql_facet


def create_output_dataset(namespace: str, metadata: dict, node_) -> list[event_v2.OutputDataset]:
    """Creates the open lineage output dataset."""
    datasource_facet = None
    storage_facet = None
    if "file_metadata" in metadata:
        name = metadata["file_metadata"]["path"]
        format = name.split(".")[-1] if "." in name else "unknown"
        storage_facet = facet_v2.storage_dataset.StorageDatasetFacet(
            storageLayer="FileSystem",
            fileFormat=format,
        )
        datasource_facet = facet_v2.datasource_dataset.DatasourceDatasetFacet(
            name=node_.name,
            uri=name,
        )
    elif "sql_metadata" in metadata:
        name = metadata["sql_metadata"]["table_name"]
    else:
        name = "--UNKNOWN--"
    schema_facet = extract_schema_facet(metadata)
    outputFacets = {}
    if storage_facet:
        outputFacets["storage"] = storage_facet
    if datasource_facet:
        outputFacets["dataSource"] = datasource_facet
    if schema_facet:
        outputFacets["schema"] = schema_facet
    if len(outputFacets) == 0:
        outputFacets = None
    outputs = [event_v2.OutputDataset(namespace, name, facets=outputFacets)]
    return outputs


class OpenLineageAdapter(
    base.BasePreGraphExecute,
    base.BasePreNodeExecute,
    base.BasePostNodeExecute,
    base.BasePostGraphExecute,
):
    """
    This adapter emits OpenLineage events.

    .. code-block:: python

        # create the openlineage client
        from openlineage.client import OpenLineageClient

        # write to file
        from openlineage.client.transport.file import FileConfig, FileTransport
        file_config = FileConfig(
            log_file_path="/path/to/your/file",
            append=False,
        )
        client = OpenLineageClient(transport=FileTransport(file_config))

        # write to HTTP, e.g. marquez
        client = OpenLineageClient(url="http://localhost:5000")

        # create the adapter
        adapter = OpenLineageAdapter(client, "my_namespace", "my_job_name")

        # add to Hamilton
        # import your pipeline code
        dr = driver.Builder().with_modules(YOUR_MODULES).with_adapters(adapter).build()
        # execute as normal -- and openlineage events will be emitted
        dr.execute(...)

    Note for data lineage to be emitted, you must use the "materializer" abstraction to provide
    metadata. See https://hamilton.dagworks.io/en/latest/concepts/materialization/.
    This can be done via the `@datasaver()` and `@dataloader()` decorators, or
    using the `@load_from` or `@save_to` decorators, as well as passing in data savers
    and data loaders via `.with_materializers()` on the Driver Builder, or via `.materialize()`
    on the driver object.
    """

    def __init__(self, client: OpenLineageClient, namespace: str, job_name: str):
        """Constructor. You pass in the OLClient.

        :param self:
        :param client:
        :param namespace:
        :param job_name:
        :return:
        """
        # self.transport = transport
        self.client = client
        self.namespace = namespace
        self.job_name = job_name

    def pre_graph_execute(
        self,
        run_id: str,
        graph: h_graph.FunctionGraph,
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """
        Emits a Run START event.
        Emits a Job Event with the sourceCode Facet for the entire DAG as the job.

        :param run_id:
        :param graph:
        :param final_vars:
        :param inputs:
        :param overrides:
        :return:
        """
        exportable_graph = graph_types.HamiltonGraph.from_graph(graph)
        graph_version = exportable_graph.version
        node_dict = [n.as_dict() for n in exportable_graph.nodes]
        job = event_v2.Job(
            namespace=self.namespace,
            name=self.job_name,
            facets={
                "sourceCode": facet_v2.source_code_job.SourceCodeJobFacet(
                    language="python",
                    sourceCode=json.dumps(node_dict),
                ),
                "jobType": facet_v2.job_type_job.JobTypeJobFacet(
                    processingType="BATCH",
                    integration="Hamilton",
                    jobType="DAG",
                ),
            },
        )
        run = event_v2.Run(
            runId=run_id,
            facets={
                "hamilton": HamiltonFacet(
                    hamilton_run_id=run_id,
                    graph_version=graph_version,
                    final_vars=final_vars,
                    inputs=list(inputs.keys()) if inputs else [],
                    overrides=list(overrides.keys()) if overrides else [],
                )
            },
        )
        run_event = event_v2.RunEvent(
            eventType=event_v2.RunState.START,
            eventTime=datetime.now(timezone.utc).isoformat(),
            run=run,
            job=job,
        )
        self.client.emit(run_event)

    def pre_node_execute(
        self, run_id: str, node_: node.Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
    ):
        """No event emitted."""
        pass

    def post_node_execute(
        self,
        run_id: str,
        node_: node.Node,
        kwargs: Dict[str, Any],
        success: bool,
        error: Optional[Exception],
        result: Optional[Any],
        task_id: Optional[str] = None,
    ):
        """
        Run Event: will emit a RUNNING event with updates on input/outputs.

        A Job Event will be emitted for graph execution, and additional SQLJob facet if data was loaded
        from a SQL source.

        A Dataset Event will be emitted if a dataloader or datasaver was used:

           - input data set if loader
           - output data set if saver
           - appropriate facets will be added to the dataset where it makes sense.

        TODO: attach statistics facets

        :param run_id:
        :param node_:
        :param kwargs:
        :param success:
        :param error:
        :param result:
        :param task_id:
        :return:
        """
        if not success:
            # do not emit anything
            return
        metadata = {}
        saved_or_loaded = ""
        if node_.tags.get("hamilton.data_saver") is True and isinstance(result, dict):
            metadata = result
            saved_or_loaded = "saved"
        elif (
            node_.tags.get("hamilton.data_loader") is True
            and node_.tags.get("hamilton.data_loader.has_metadata") is True
            and isinstance(result, tuple)
            and len(result) == 2
            and isinstance(result[1], dict)
        ):
            metadata = result[1]
            saved_or_loaded = "loaded"
        if not metadata:
            # no metadata to emit
            return

        inputs = []
        outputs = []
        sql_facet = None
        if saved_or_loaded == "loaded":
            inputs, sql_facet = create_input_dataset(self.namespace, metadata, node_)
        else:
            outputs = create_output_dataset(self.namespace, metadata, node_)

        run = event_v2.Run(
            runId=run_id,
        )
        job_facets = {}
        if sql_facet:
            job_facets["sql"] = sql_facet
        job = event_v2.Job(namespace=self.namespace, name=self.job_name, facets=job_facets)
        run_event = event_v2.RunEvent(
            eventType=event_v2.RunState.RUNNING,
            eventTime=datetime.now(timezone.utc).isoformat(),
            run=run,
            job=job,
            inputs=inputs,
            outputs=outputs,
        )
        self.client.emit(run_event)

    def post_graph_execute(
        self,
        run_id: str,
        graph: h_graph.FunctionGraph,
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
    ):
        """Emits a Run COMPLETE or FAIL event.

        :param run_id:
        :param graph:
        :param success:
        :param error:
        :param results:
        :return:
        """
        job = event_v2.Job(
            namespace=self.namespace,
            name=self.job_name,
        )
        facets = {}
        run_event_type = event_v2.RunState.COMPLETE
        if error:
            run_event_type = event_v2.RunState.FAIL
            error_message = str(error)
            facets = {
                "errorMessage": facet_v2.error_message_run.ErrorMessageRunFacet(
                    message=error_message,
                    stackTrace=get_stack_trace(error),
                    programmingLanguage="python",
                )
            }
        run = event_v2.Run(runId=run_id, facets=facets)

        run_event = event_v2.RunEvent(
            eventType=run_event_type,
            eventTime=datetime.now(timezone.utc).isoformat(),
            run=run,
            job=job,
        )
        self.client.emit(run_event)


# if __name__ == "__main__":
#     from openlineage.client import OpenLineageClient
#     from openlineage.client.transport.file import FileConfig, FileTransport
#
#     file_config = FileConfig(
#         log_file_path="/path/to/your/file",
#         append=False,
#     )
#
#     client = OpenLineageClient(transport=FileTransport(file_config))
#     namespace = "my_namespace"
#     db_datset = Dataset(namespace, name, facets)



import dataclasses
import pathlib
from typing import IO, Any, Collection, Dict, List, Optional, Type, Union

try:
    import plotly.graph_objects
except ImportError as e:
    raise NotImplementedError("Plotly is not installed.") from e

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataSaver


@dataclasses.dataclass
class PlotlyStaticWriter(DataSaver):
    """Write Plotly figure as static image format
    ref: https://plotly.com/python/static-image-export/
    """

    path: Union[str, pathlib.Path, IO]
    format: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    scale: Optional[Union[int, float]] = None
    validate: bool = True
    engine: str = "auto"

    def _get_saving_kwargs(self) -> dict:
        kwargs = {}
        if self.format is not None:
            kwargs["format"] = self.format
        if self.width is not None:
            kwargs["width"] = self.width
        if self.height is not None:
            kwargs["height"] = self.height
        if self.scale is not None:
            kwargs["scale"] = self.scale
        if self.validate is not None:
            kwargs["validate"] = self.validate
        if self.engine is not None:
            kwargs["engine"] = self.engine

        return kwargs

    def save_data(self, data: plotly.graph_objects.Figure) -> Dict[str, Any]:
        data.write_image(file=self.path, **self._get_saving_kwargs())
        return utils.get_file_metadata(self.path)

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [plotly.graph_objects.Figure]

    @classmethod
    def name(cls) -> str:
        return "plotly"


@dataclasses.dataclass
class PlotlyInteractiveWriter(DataSaver):
    """Write Plotly figure as interactive HTML + JS
    ref: https://plotly.com/python/interactive-html-export/
    """

    path: Union[str, pathlib.Path, IO]
    config: Optional[Dict] = None
    auto_play: bool = True
    include_plotlyjs: Union[bool, str] = (
        True  # or "cdn", "directory", "require", "False", "other string .js"
    )
    include_mathjax: Union[bool, str] = False  # "cdn", "string .js"
    post_script: Union[str, List[str], None] = None
    full_html: bool = True
    animation_opts: Optional[Dict] = None
    default_width: Union[int, float, str] = "100%"
    default_height: Union[int, float, str] = "100%"
    validate: bool = True
    auto_open: bool = True
    div_id: Optional[str] = None

    def _get_saving_kwargs(self) -> dict:
        kwargs = {}
        if self.config is not None:
            kwargs["config"] = self.config
        if self.auto_play is not None:
            kwargs["auto_play"] = self.auto_play
        if self.include_plotlyjs is not None:
            kwargs["include_plotlyjs"] = self.include_plotlyjs
        if self.include_mathjax is not None:
            kwargs["include_mathjax"] = self.include_mathjax
        if self.post_script is not None:
            kwargs["post_script"] = self.post_script
        if self.full_html is not None:
            kwargs["full_html"] = self.full_html
        if self.animation_opts is not None:
            kwargs["animation_opts"] = self.animation_opts
        if self.default_width is not None:
            kwargs["default_width"] = self.default_width
        if self.default_height is not None:
            kwargs["default_height"] = self.default_height
        if self.validate is not None:
            kwargs["validate"] = self.validate
        if self.auto_open is not None:
            kwargs["auto_open"] = self.auto_open
        if self.div_id is not None:
            kwargs["div_id"] = self.div_id
        return kwargs

    def save_data(self, data: plotly.graph_objects.Figure) -> Dict[str, Any]:
        data.write_html(file=self.path, **self._get_saving_kwargs())
        return utils.get_file_metadata(self.path)

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [plotly.graph_objects.Figure]

    @classmethod
    def name(cls) -> str:
        return "html"


def register_data_loaders():
    for loader in [
        PlotlyStaticWriter,
        PlotlyInteractiveWriter,
    ]:
        registry.register_adapter(loader)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False



from unittest import mock

import pandas as pd
import pytest

from hamilton import base, node
from hamilton.caching.adapter import HamiltonCacheAdapter
from hamilton.driver import (
    Builder,
    Driver,
    InvalidExecutorException,
    TaskBasedGraphExecutor,
    Variable,
)
from hamilton.execution import executors
from hamilton.io.materialization import from_, to

import tests.resources.cyclic_functions
import tests.resources.dummy_functions
import tests.resources.dynamic_parallelism.parallel_linear_basic
import tests.resources.tagging
import tests.resources.test_default_args
import tests.resources.test_driver_serde_mapper
import tests.resources.test_driver_serde_worker
import tests.resources.test_for_materialization
import tests.resources.very_simple_dag

"""This file tests driver capabilities.
Anything involving execution is tested for multiple executors/driver configuration.
Anything not involving execution is tested for just the single driver configuration.

TODO -- move any execution tests to tests the graph executor capabilities on their own.
"""


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({"a": 1})),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"a": 1})
            .build()
        ),
    ],
)
def test_driver_validate_input_types(driver_factory):
    dr = driver_factory()
    results = dr.raw_execute(["a"])
    assert results == {"a": 1}


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.very_simple_dag)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.very_simple_dag)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
)
def test_driver_validate_runtime_input_types(driver_factory):
    dr = driver_factory()
    results = dr.raw_execute(["b"], inputs={"a": 1})
    assert results == {"b": 1}


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.cyclic_functions)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.cyclic_functions)
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
)
def test_driver_has_cycles_true(driver_factory):
    """Tests that we don't break when detecting cycles from the driver."""
    dr = driver_factory()
    assert dr.has_cycles(["C"])


# This is possible -- but we don't want to officially support it. Here for documentation purposes.
# def test_driver_cycles_execute_override():
#     """Tests that we short circuit a cycle by passing in overrides."""
#     dr = Driver({}, tests.resources.cyclic_functions, adapter=base.DefaultAdapter())
#     result = dr.execute(['C'], overrides={'D': 1}, inputs={'b': 2, 'c': 2})
#     assert result['C'] == 34


@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.cyclic_functions)),
        # TODO -- fix erroring out when we try to run a driver with cycles
        # should display a better error
        # (lambda: Builder()
        #     .enable_parallelizable_type(allow_experimental_mode=True)
        #     .with_modules(tests.resources.cyclic_functions)
        #     .with_remote_executor(executors.SynchronousLocalTaskExecutor())
        #     .with_adapter(base.DefaultAdapter())
        #     .build())
    ],
)
def test_driver_cycles_execute_recursion_error(driver_factory):
    """Tests that we throw a recursion error when we try to execute over a DAG that isn't a DAG."""
    dr = driver_factory()
    with pytest.raises(RecursionError):
        dr.execute(["C"], inputs={"b": 2, "c": 2})


def test_driver_variables_exposes_tags():
    dr = Driver({}, tests.resources.tagging)
    tags = {var.name: var.tags for var in dr.list_available_variables()}
    assert tags["a"] == {"module": "tests.resources.tagging", "test": "a"}
    assert tags["b"] == {"module": "tests.resources.tagging", "test": "b_c"}
    assert tags["c"] == {"module": "tests.resources.tagging", "test": "b_c"}
    assert tags["d"] == {"module": "tests.resources.tagging", "test_list": ["us", "uk"]}


@pytest.mark.parametrize(
    "filter,expected",
    [
        (None, {"a", "b_c", "b", "c", "d"}),  # no filter
        ({}, {"a", "b_c", "b", "c", "d"}),  # empty filter
        ({"test": "b_c"}, {"b", "c"}),
        ({"test": None}, {"a", "b", "c"}),
        ({"module": "tests.resources.tagging"}, {"a", "b_c", "b", "c", "d"}),
        ({"test_list": "us"}, {"d"}),
        ({"test_list": "uk"}, {"d"}),
        ({"test_list": ["uk"]}, {"d"}),
        ({"module": "tests.resources.tagging", "test": "b_c"}, {"b", "c"}),
        ({"test_list": ["nz", "uk"]}, {"d"}),
        ({"test_list": ["us", "uk"]}, {"d"}),
        ({"test_list": ["uk", "us"]}, {"d"}),
        ({"test": ["b_c"]}, {"b", "c"}),
        ({"test": ["b_c", "foo"]}, {"b", "c"}),
    ],
    ids=[
        "filter with None passed",
        "filter with empty filter",
        "filter by single tag with extract decorator",
        "filter with None value",
        "filter with specific value",
        "filter tag with list values - value 1",
        "filter tag with list values - value 2",
        "filter tag with list values - query is single node list",
        "filter with two filter clauses",
        "filter with with list values not exact OR interpretation",
        "filter with with list values exact",
        "filter with with list values exact order invariant",
        "filter with with list values edge case one item match",
        "filter with with list values OR interpretation",
    ],
)
def test_driver_variables_filters_tags(filter, expected):
    dr = Driver({}, tests.resources.tagging)
    actual = {var.name for var in dr.list_available_variables(tag_filter=filter)}
    assert actual == expected


def test_driver_variables_filters_tags_error():
    dr = Driver({}, tests.resources.tagging)
    with pytest.raises(ValueError):
        # non string value is not allowed
        dr.list_available_variables(tag_filter={"test": 1234})
    with pytest.raises(ValueError):
        # empty list shouldn't be allowed
        dr.list_available_variables(tag_filter={"test": []})


def test_driver_variables_external_input():
    dr = Driver({}, tests.resources.very_simple_dag)
    input_types = {var.name: var.is_external_input for var in dr.list_available_variables()}
    assert input_types["a"] is True
    assert input_types["b"] is False


def test_driver_variables_exposes_original_function():
    dr = Driver({}, tests.resources.very_simple_dag)
    originating_functions = {
        var.name: var.originating_functions for var in dr.list_available_variables()
    }
    assert originating_functions["b"] == (tests.resources.very_simple_dag.b,)
    assert originating_functions["a"] == (tests.resources.very_simple_dag.b,)  # a is an input


@mock.patch("hamilton.telemetry.send_event_json")
def test_capture_constructor_telemetry_disabled(send_event_json):
    """Tests that we don't do anything if telemetry is disabled."""
    send_event_json.return_value = ""
    Driver({}, tests.resources.tagging)  # this will exercise things underneath.
    assert send_event_json.called is False


@mock.patch("hamilton.telemetry.get_adapter_name")
@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
def test_capture_constructor_telemetry_error(send_event_json, get_adapter_name):
    """Tests that we don't error if an exception occurs"""
    get_adapter_name.side_effect = ValueError("TELEMETRY ERROR")
    Driver({}, tests.resources.tagging)  # this will exercise things underneath.
    assert send_event_json.called is False


@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
def test_capture_constructor_telemetry_none_values(send_event_json):
    """Tests that we don't error if there are none values"""
    Driver({}, None, None)  # this will exercise things underneath.
    assert send_event_json.called is True


@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
def test_capture_constructor_telemetry(send_event_json):
    """Tests that we send an event if we could. Validates deterministic parts."""
    Driver({}, tests.resources.very_simple_dag)
    # assert send_event_json.called is True
    assert len(send_event_json.call_args_list) == 1  # only called once
    # check contents of what it was called with:
    send_event_json_call = send_event_json.call_args_list[0]
    actual_event_dict = send_event_json_call[0][0]
    assert actual_event_dict["api_key"] == "phc_mZg8bkn3yvMxqvZKRlMlxjekFU5DFDdcdAsijJ2EH5e"
    assert actual_event_dict["event"] == "os_hamilton_run_start"
    # validate schema
    expected_properties = {
        "os_type",
        "os_version",
        "python_version",
        "distinct_id",
        "hamilton_version",
        "telemetry_version",
        "number_of_nodes",
        "number_of_modules",
        "number_of_config_items",
        "decorators_used",
        "graph_adapter_used",
        "result_builder_used",
        "driver_run_id",
        "error",
        "graph_executor_class",
        "lifecycle_adapters_used",
    }
    actual_properties = actual_event_dict["properties"]
    assert set(actual_properties.keys()) == expected_properties
    # validate static parts
    assert actual_properties["error"] is None
    assert actual_properties["number_of_nodes"] == 2  # b, and input a
    assert actual_properties["number_of_modules"] == 1
    assert actual_properties["number_of_config_items"] == 0
    assert actual_properties["number_of_config_items"] == 0
    assert actual_properties["graph_adapter_used"] == "deprecated -- see lifecycle_adapters_used"
    assert actual_properties["result_builder_used"] == "hamilton.base.PandasDataFrameResult"
    assert actual_properties["lifecycle_adapters_used"] == ["hamilton.base.PandasDataFrameResult"]


@mock.patch("hamilton.telemetry.send_event_json")
@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.very_simple_dag)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.very_simple_dag)
            .with_adapter(base.SimplePythonGraphAdapter(base.PandasDataFrameResult()))
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
)
def test_capture_execute_telemetry_disabled(send_event_json, driver_factory):
    """Tests that we don't do anything if telemetry is disabled."""
    dr = driver_factory()
    results = dr.execute(["b"], inputs={"a": 1})
    expected = pd.DataFrame([{"b": 1}])
    pd.testing.assert_frame_equal(results, expected)
    assert send_event_json.called is False


@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.very_simple_dag)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.very_simple_dag)
            .with_adapter(base.SimplePythonGraphAdapter(base.PandasDataFrameResult()))
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
)
def test_capture_execute_telemetry_error(send_event_json, driver_factory):
    """Tests that we don't error if an exception occurs"""
    send_event_json.side_effect = [None, ValueError("FAKE ERROR"), None]
    dr = driver_factory()
    results = dr.execute(["b"], inputs={"a": 1})
    expected = pd.DataFrame([{"b": 1}])
    pd.testing.assert_frame_equal(results, expected)
    assert send_event_json.called is True
    assert len(send_event_json.call_args_list) == 2


@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({}, tests.resources.very_simple_dag)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.very_simple_dag)
            .with_adapter(base.SimplePythonGraphAdapter(base.PandasDataFrameResult()))
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .build()
        ),
    ],
)
def test_capture_execute_telemetry(send_event_json, driver_factory):
    """Happy path with values passed."""
    dr = driver_factory()
    results = dr.execute(["b"], inputs={"a": 1}, overrides={"b": 2})
    expected = pd.DataFrame([{"b": 2}])
    pd.testing.assert_frame_equal(results, expected)
    assert send_event_json.called is True
    assert len(send_event_json.call_args_list) == 2


@mock.patch("hamilton.telemetry.send_event_json")
@mock.patch("hamilton.telemetry.g_telemetry_enabled", True)
@pytest.mark.parametrize(
    "driver_factory",
    [
        (lambda: Driver({"a": 1}, tests.resources.very_simple_dag)),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.very_simple_dag)
            .with_adapter(base.SimplePythonGraphAdapter(base.PandasDataFrameResult()))
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"a": 1})
            .build()
        ),
    ],
)
def test_capture_execute_telemetry_none_values(send_event_json, driver_factory):
    """Happy path with none values."""
    dr = driver_factory()
    results = dr.execute(["b"])
    expected = pd.DataFrame([{"b": 1}])
    pd.testing.assert_frame_equal(results, expected)
    assert len(send_event_json.call_args_list) == 2


@pytest.mark.parametrize(
    "driver_factory",
    [
        (
            lambda: Driver(
                {"required": 1},
                tests.resources.test_default_args,
                adapter=base.DefaultAdapter(),
            )
        ),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.test_default_args)
            .with_adapter(base.DefaultAdapter())
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"required": 1})
            .build()
        ),
    ],
)
def test_node_is_required_by_anything(driver_factory):
    """Tests that default args are correctly interpreted.

    Specifically, if it's not in the execution path then things should
    just work. Here I'm being lazy and rather than specifically testing
    _node_is_required_by_anything() directly, I'm doing it via
    execute(), which calls it via validate_inputs().

    To understand what's going on see the functions in `test_default_args`.
    """
    dr = driver_factory()
    # D is not in the execution path, but requires defaults_to_zero
    # so this should work.
    results = dr.execute(["C"])
    assert results["C"] == 2
    with pytest.raises(ValueError):
        # D is now in the execution path, but requires defaults_to_zero
        # this should error
        dr.execute(["D"])


@pytest.mark.parametrize(
    "driver_factory",
    [
        (
            lambda: Driver(
                {"required": 1},
                tests.resources.test_default_args,
                adapter=base.DefaultAdapter(),
            )
        ),
        (
            lambda: Builder()
            .enable_dynamic_execution(allow_experimental_mode=True)
            .with_modules(tests.resources.test_default_args)
            .with_adapter(base.DefaultAdapter())
            .with_remote_executor(executors.SynchronousLocalTaskExecutor())
            .with_config({"required": 1})
            .build()
        ),
    ],
)
def test_using_callables_to_execute(driver_factory):
    """Test that you can pass a function reference and it will work fine."""
    dr = driver_factory()
    results = dr.execute(
        [tests.resources.test_default_args.C, tests.resources.test_default_args.B, "A"]
    )
    assert results["C"] == 2
    assert results["B"] == 1
    assert results["A"] == 1
    with pytest.raises(ValueError):
        dr.execute([tests.resources.cyclic_functions.B])


def test_create_final_vars():
    """Tests that the final vars are created correctly."""
    dr = Driver({"required": 1}, tests.resources.test_default_args)
    D_node = dr.graph.nodes["D"]
    actual = dr._create_final_vars(
        [
            "C",
            tests.resources.test_default_args.B,
            tests.resources.test_default_args.A,
            Variable.from_node(D_node),
        ]
    )
    expected = ["C", "B", "A", "D"]
    assert actual == expected


def test_create_final_vars_errors():
    """Tests that we catch functions pointed to in modules that aren't part of the DAG."""
    dr = Driver({"required": 1}, tests.resources.test_default_args)
    with pytest.raises(ValueError):
        dr._create_final_vars(
            ["C", tests.resources.cyclic_functions.A, tests.resources.cyclic_functions.B]
        )


def test_v2_driver_builder():
    dr = (
        Builder()
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_adapter(base.DefaultAdapter())
        .with_modules(tests.resources.very_simple_dag)
        .build()
    )
    assert isinstance(dr.graph_executor, TaskBasedGraphExecutor)
    assert list(dr.graph_modules) == [tests.resources.very_simple_dag]


def test_executor_validates_happy_default_executor():
    dr = Driver({}, tests.resources.very_simple_dag)
    nodes, user_nodes = dr.graph.get_upstream_nodes(["b"])
    dr.graph_executor.validate(nodes | user_nodes)


def test_executor_validates_sad_default_executor():
    dr = Driver({}, tests.resources.dynamic_parallelism.parallel_linear_basic)
    nodes, user_nodes = dr.graph.get_upstream_nodes(["final"])
    with pytest.raises(InvalidExecutorException):
        dr.graph_executor.validate(nodes | user_nodes)


def test_executor_validates_happy_parallel_executor():
    dr = (
        Builder()
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_modules(tests.resources.dynamic_parallelism.parallel_linear_basic)
        .build()
    )

    nodes, user_nodes = dr.graph.get_upstream_nodes(["final"])
    dr.graph_executor.validate(nodes | user_nodes)


def test_builder_defaults_to_dict_result():
    dr = Builder().with_modules(tests.resources.dummy_functions).build()

    result = dr.execute(["C"], inputs={"b": 1, "c": 1})
    assert result == {"C": 4}


def test_builder_copy():
    builder = (
        Builder()
        .with_modules(tests.resources.dummy_functions)
        .with_config({"config_key": 13})
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_adapter(base.DefaultAdapter())
        .with_local_executor(executors.SynchronousLocalTaskExecutor())
        .with_remote_executor(executors.SynchronousLocalTaskExecutor())
    )
    builder_copy = builder.copy()

    assert builder_copy is not builder
    for attr, attr_value in builder.__dict__.items():
        attr_value_copy = getattr(builder_copy, attr)
        assert attr_value_copy == attr_value
        # TODO check that each objects
        # if isinstance(attr_value, bool):
        #     continue
        # assert attr_value_copy is not attr_value


def test_builder_with_loader_materializer():
    loader_target = "external"
    loader = from_.json(target=loader_target, path="/my/file.json")
    dr = (
        Builder()
        .with_modules(tests.resources.test_for_materialization)
        .with_materializers(loader)
        .build()
    )

    assert any(n.name == f"load_data.{loader_target}" for n in dr.graph.get_nodes())


def test_builder_with_saver_materializer():
    saver_id = "saver_node"
    saver = to.json(
        id=saver_id,
        dependencies=["expects_loader"],
        path="/my/file.json",
    )
    dr = (
        Builder()
        .with_modules(tests.resources.test_for_materialization)
        .with_materializers(saver)
        .build()
    )

    assert any(n.name == saver_id for n in dr.graph.get_nodes())


def test_builder_materializer_and_execution_materializer(tmp_path):
    static_saver = to.json(
        id="static_saver",
        dependencies=["json_to_save_1"],
        path=f"{tmp_path}/file.json",
    )
    dynamic_saver = to.json(
        id="dynamic_saver",
        dependencies=["json_to_save_2"],
        path=f"{tmp_path}/file2.json",
    )
    dr = (
        Builder()
        .with_modules(tests.resources.test_for_materialization)
        .with_materializers(static_saver)
        .build()
    )
    metadata, additional = dr.materialize(dynamic_saver, additional_vars=["static_saver"])

    assert "dynamic_saver" in metadata.keys()
    assert "static_saver" in additional.keys()


def test_builder_materializer_error():
    static_saver = to.json(
        id="static_saver",
        dependencies=["json_to_save_1"],
        path="/file.json",
    )

    materializers = [static_saver]
    with pytest.raises(ValueError):
        (
            Builder()
            .with_modules(tests.resources.test_for_materialization)
            .with_materializers(materializers)
            .build()
        )

    # also check that using `*` leads to no issue
    (
        Builder()
        .with_modules(tests.resources.test_for_materialization)
        .with_materializers(*materializers)
        .build()
    )


def test_materialize_checks_required_input(tmp_path):
    dr = Builder().with_modules(tests.resources.dummy_functions).build()

    with pytest.raises(ValueError):
        dr.materialize(additional_vars=["C"], inputs={"c": 1})
    with pytest.raises(ValueError):
        dr.materialize(
            to.pickle(id="1", path=f"{tmp_path}/foo.pkl", dependencies=["C"]), inputs={"c": 1}
        )


def test_cache_raise_if_setting_twice(tmp_path):
    builder = Builder()

    builder.with_cache(path=tmp_path)
    # case 1: .with_cache() then .with_cache()
    with pytest.raises(ValueError):
        builder.with_cache(path=tmp_path)
    # case 2: .with_cache() then adding SmartCacheAdapter()
    with pytest.raises(ValueError):
        builder.with_adapters(HamiltonCacheAdapter(path=tmp_path))
    # case 3: add SmartCacheAdapter() then .with_cache()
    builder = Builder()
    builder.with_adapters(HamiltonCacheAdapter(path=tmp_path))
    with pytest.raises(ValueError):
        builder.with_cache()


def test_validate_execution_happy():
    dr = Builder().with_modules(tests.resources.very_simple_dag).build()
    dr.validate_execution(["b"], inputs={"a": 1})


def test_validate_execution_sad():
    dr = Builder().with_modules(tests.resources.very_simple_dag).build()
    with pytest.raises(ValueError):
        dr.validate_execution(["b"], inputs={})


def test_validate_materialization_happy(tmp_path):
    dr = Builder().with_modules(tests.resources.very_simple_dag).build()
    dr.validate_materialization(
        to.pickle(id="1", path=f"{tmp_path}/foo.pkl", dependencies=["b"]), inputs={"a": 1}
    )


def test_validate_materialization_sad(tmp_path):
    dr = Builder().with_modules(tests.resources.very_simple_dag).build()
    with pytest.raises(ValueError):
        dr.validate_materialization(
            # c does not exist
            # no inputs either
            to.pickle(id="1", path=f"{tmp_path}/foo.pkl", dependencies=["c"]),
            inputs={},
        )


def test_variable_from_node():
    # Quick test for creating variables from nodes --
    # this is simple but its nice to have

    def func_to_test(a: int) -> int:
        """This is a doctstring"""
        return a + 1

    n = node.Node.from_fn(func_to_test)
    v = Variable.from_node(n)
    assert v.name == n.name
    assert v.type == n.type
    assert v.tags == n.tags
    assert v.documentation == n.documentation == "This is a doctstring"
    assert v.originating_functions == n.originating_functions


def test_driver_setstate_getstate():
    """This is an integration test testing serializability of the hamilton driver."""
    from hamilton.execution import executors

    drivers = []
    inputs = []
    for i in range(4):
        dr = Builder().with_modules(tests.resources.test_driver_serde_worker).build()
        drivers.append(dr)
        inputs.append({"a": i})

    dr = (
        Builder()
        .with_modules(tests.resources.test_driver_serde_mapper)
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_remote_executor(executors.MultiProcessingExecutor(4))
        .build()
    )
    r = dr.execute(
        final_vars=["reducer"],
        inputs={"drivers": drivers, "inputs": inputs, "final_vars": ["double"]},
    )
    assert r == {"reducer": [{"double": 0}, {"double": 2}, {"double": 4}, {"double": 6}]}



import pandas as pd

from hamilton import htypes


def base_func(a: pd.Series, b: pd.Series) -> htypes.column[pd.Series, int]:
    """Pandas UDF function."""
    return a + b


def base_func2(base_func: int, c: int) -> int:
    """Vanilla UDF function. This depends on a pandas UDF function."""
    return base_func + c


def base_func3(c: int, d: int) -> int:
    """This function is not satisfied by the dataframe. So is computed without using
    the columns in the dataframe."""
    return c + d
